unit FormsMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, OleServer, MSXML_TLB, ActnList, Menus,
  ExtCtrls, Contnrs, ComCtrls;

type
  TRoad = class;

  TTown = class
  private
    FName: string;
    FX: Integer;
    FY: Integer;
  public
    // pro potreby Dijkstry...
    TownDist: Double;
    DijkState: (dsTemp, dsFinal);
    FromTown: TTown;
    Road: TRoad;

    property Name: string read FName write FName;
    property x: Integer read FX write FX;
    property Y: Integer read FY write FY;
  end;

  TRoad = class
  private
    function GetCenter: TPoint;
  public
    tFrom: TTown;
    tTo: TTown;
    Length: Integer;
    Velocity: Integer;

    Highlight: Boolean;

    property Center: TPoint read GetCenter;
  end;

  TSpotreba = class
  public
    Velocity: Integer;
    Litru: Double;
  end;

  TMap = class
  private
    FTowns: TObjectList;
    FRoads: TObjectList;
    FSpotreby: TObjectList;
  public
    constructor Create(Node: IXMLDOMNode);

    property Towns: TObjectList read FTowns;
    property Roads: TObjectList read FRoads;
    property Spotreby: TObjectList read FSpotreby;

    function FindTown(Town: string): TTown;
    function GetSpotreba(AVelocity: Integer): Double;
  end;

  TfrmMain = class(TForm)
    xmlMap: TDOMDocument;
    mmnMain: TMainMenu;
    mniMenuMap: TMenuItem;
    mniMapLoad: TMenuItem;
    aclMain: TActionList;
    actMapLoad: TAction;
    opdMap: TOpenDialog;
    ptbMap: TPaintBox;
    scbHoriz: TScrollBar;
    scbVert: TScrollBar;
    mniViewMenu: TMenuItem;
    mniViewRoads: TMenuItem;
    actViewRoads: TAction;
    trbZoom: TTrackBar;
    stbMain: TStatusBar;
    Label1: TLabel;
    Kriteria1: TMenuItem;
    mniKritLength: TMenuItem;
    mniKritVelocity: TMenuItem;
    mniKritCost: TMenuItem;
    actSavePath: TAction;
    mniSavePath: TMenuItem;
    svdPath: TSaveDialog;
    pnlInfopanel: TPanel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    lblSelection: TLabel;
    lblLength: TLabel;
    lblTime: TLabel;
    lblLitru: TLabel;
    actViewInfopanel: TAction;
    Infopanel1: TMenuItem;
    procedure actMapLoadExecute(Sender: TObject);
    procedure ptbMapPaint(Sender: TObject);
    procedure scbHorizScroll(Sender: TObject; ScrollCode: TScrollCode;
      var ScrollPos: Integer);
    procedure actViewRoadsExecute(Sender: TObject);
    procedure trbZoomChange(Sender: TObject);
    procedure ptbMapMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ptbMapMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure mniKritLengthClick(Sender: TObject);
    procedure actSavePathExecute(Sender: TObject);
    procedure actSavePathUpdate(Sender: TObject);
    procedure actViewInfopanelExecute(Sender: TObject);
  private
    Map: TMap;
    TownLeft, TownTop, TownBottom, TownRight: TTown;
    FSelectedTown: TTown;
    FSelectedRoad: TRoad;
    PathFrom, PathTo: TTown;
    Len, Time, palivo: Double;
    procedure SetSelectedTown(Value: TTown);
    procedure SetSelectedRoad(Value: TRoad);
  public
    property SelectedTown: TTown read FSelectedTown write SetSelectedTown;
    property SelectedRoad: TRoad read FSelectedRoad write SetSelectedRoad;

    procedure NajdiCestu(tFrom, tTo: TTown);
  end;

var
  frmMain: TfrmMain;

implementation

{$R *.dfm}

function TMap.GetSpotreba(AVelocity: Integer): Double;
var
  I: Integer;
begin
  for I := 0 to Spotreby.Count - 1 do
    with Spotreby[I] as TSpotreba do
      if Velocity = AVelocity then begin
        Result := Litru;
        Exit;
      end;
  Result := 0;
end;


procedure TfrmMain.SetSelectedTown(Value: TTown);
begin
  FSelectedRoad := nil;
  FSelectedTown := Value;
  ptbMap.Invalidate;
  stbMain.SimpleText := Format('Vybrno msto: %s', [Value.Name]);
end;

procedure TfrmMain.SetSelectedRoad(Value: TRoad);
begin
  FSelectedTown := nil;
  FSelectedRoad := Value;
  ptbMap.Invalidate;
  stbMain.SimpleText := Format('Vybrna cesta z %s do %s, dlka %d, rychlost %d, doba jizdy %s',
    [Value.tFrom.Name, Value.tTo.Name, Value.Length, Value.Velocity, TimeToStr(Value.Length / Value.Velocity / 24)]);

  lblSelection.Caption := Format('Z %s do %s', [Value.tFrom.Name, Value.tTo.Name]);
  lblLength.Caption := IntToStr(Value.Length);
  lblTime.Caption := TimeToStr(Value.Length / Value.Velocity / 24);
  lblLitru.Caption := FloatToStr(Map.GetSpotreba(Value.Velocity) / 100 * Value.Length);
end;

function TRoad.GetCenter: TPoint;
begin
  Result.X := tTo.x + (tFrom.x - tTo.x) div 2;
  Result.Y := tTo.Y + (tFrom.Y - tTo.Y) div 2;
end;

constructor TMap.Create(Node: IXMLDOMNode);

  procedure LoadTown(Town: IXMLDOMNode);
  var
    I: Integer;
    T: TTown;
  begin
    T := TTown.Create;
    for I := 0 to Town.childNodes.length - 1 do
      if Town.childNodes[I].nodeName = 'nazev' then
        T.Name := Town.childNodes[I].text
      else if Town.childNodes[I].nodeName = 'polohaX' then
        T.x := StrToInt(Town.childNodes[I].text)
      else if Town.childNodes[I].nodeName = 'polohaY' then
        T.Y := StrToInt(Town.childNodes[I].text);
    Towns.Add(T);
  end;

  procedure LoadRoad(Road: IXMLDOMNode);
  var
    R: TRoad;
    I: Integer;
  begin
    R := TRoad.Create;
    for I := 0 to Road.childNodes.length - 1 do
      if Road.childNodes[I].nodeName = 'z' then
        R.tFrom := FindTown(Road.childNodes[I].text)
      else if Road.childNodes[I].nodeName = 'do' then
        R.tTo := FindTown(Road.childNodes[I].text)
      else if Road.childNodes[I].nodeName = 'vzdalenost' then
        R.Length := StrToInt(Road.childNodes[I].text)
      else if Road.childNodes[I].nodeName = 'prumernaRychlost' then
        R.Velocity := StrToInt(Road.childNodes[I].text);
    Roads.Add(R);
  end;

  procedure LoadSpotreba(Spotreba: IXMLDOMNode);
  var
    S: TSpotreba;
    I: Integer;
    FormatSettings: TFormatSettings;
  begin
    S := TSpotreba.Create;
    for I := 0 to Spotreba.childNodes.length - 1 do
      if Spotreba.childNodes[I].nodeName = 'rychlost' then
        S.Velocity := StrToInt(Spotreba.childNodes[I].text)
      else if Spotreba.childNodes[I].nodeName = 'litru' then begin
        FormatSettings.DecimalSeparator := '.';
        S.Litru := StrToFloat(Spotreba.childNodes[I].text, FormatSettings);
      end;
    Spotreby.Add(S);
  end;

var
  I, J: Integer;
begin
  FTowns := TObjectList.Create(True);
  FRoads := TObjectList.Create(True);
  FSpotreby := TObjectList.Create(True);

  for I := 0 to Node.childNodes.length - 1 do
    if Node.childNodes.item[I].nodeName = 'mesta' then
      for J := 0 to Node.childNodes.item[I].childNodes.length - 1 do
        if Node.childNodes[I].childNodes[J].nodeName = 'mesto' then
          LoadTown(Node.childNodes[I].childNodes[J])
        else
    else if Node.childNodes.item[I].nodeName = 'cesty' then
      for J := 0 to Node.childNodes.item[I].childNodes.length - 1 do
        if Node.childNodes[I].childNodes[J].nodeName = 'cesta' then
          LoadRoad(Node.childNodes[I].childNodes[J])
        else
    else if Node.childNodes.item[I].nodeName = 'spotreby' then
      for J := 0 to Node.childNodes.item[I].childNodes.length - 1 do
        if Node.childNodes[I].childNodes[J].nodeName = 'spotreba' then
          LoadSpotreba(Node.childNodes[I].childNodes[J]);
end;

function TMap.FindTown(Town: string): TTown;
var
  I: Integer;
begin
  for I := 0 to Towns.Count - 1 do
    if (Towns[I] as TTown).Name = Town then begin
      Result := Towns[I] as TTown;
      Exit;
    end;
  Result := nil;
end;

procedure TfrmMain.actMapLoadExecute(Sender: TObject);
var
  I: Integer;
begin
  if opdMap.Execute then
    if not xmlMap.load(opdMap.FileName) then
      raise Exception.Create('Nelze nast vstupn soubor');
  if Assigned(Map) then begin
    Map.Free;
    Map := nil;
  end;
  PathFrom := nil;
  PathTo := nil;
  FSelectedTown := nil;
  FSelectedRoad := nil;

  Map := TMap.Create(xmlMap.documentElement);

{// Zrusime vsechna aktivni mesta...
  for I := ComponentCount - 1 downto 0 do
    if Components[I].Tag = 25 then
      Components[I].Free;  }

// Nalezneme krajni mesta
  TownLeft := nil; TownTop := nil; TownBottom := nil; TownRight := nil;
  for I := 0 to Map.Towns.Count - 1 do
    with Map.Towns[I] as TTown do begin
      if (TownLeft = nil) or (TownLeft.x > x) then TownLeft := Map.Towns[I] as TTown;
      if (TownRight  = nil) or (TownRight .x < x) then TownRight  := Map.towns[I] as TTown;
      if (TownTop = nil) or (TownTop.Y > Y) then TownTop := Map.towns[I] as TTown;
      if (TownBottom = nil) or (TownBottom.Y < Y) then TownBottom := Map.towns[I] as TTown;
    end;

  scbHoriz.Min := TownLeft.x;
  scbHoriz.Max := TownRight.x;
  scbHoriz.Position := (Townleft.X + Townright.X) div 2;
  scbVert.Min := TownTop.Y;
  scbVert.Max := TownBottom.Y;
  scbVert.Position := (TownTop.Y + TownBottom.Y) div 2;

  ptbMap.Invalidate;

{// Vytvorime nova mesta
  for I := 0 to Map.Towns.Count - 1 do
    with Map.Towns[I] as TTown do begin
      Shape := TShape.Create(Self);
      Shape.Parent := Self;
      Shape.Top := Y - scbVert.Position + ClientHeight div 2;
      Shape.Left := X - scbVert.Position + ClientWidth div 2;
      Shape.Height := 10;
      Shape.Width := 10;
      Shape.Shape := stCircle;
      Shape.Tag := 25;
    end;    }
end;

procedure TfrmMain.ptbMapPaint(Sender: TObject);
var
  I: Integer;

  function CalcX(X: Integer): Integer;
  begin
    Result := Round((X - scbHoriz.Position) *(trbZoom.Position +25 ) / 50) + ClientWidth div 2;
  end;

  function CalcY(Y: Integer): Integer;
  begin
    Result := Round((Y - scbVert.Position) * (trbZoom.Position +25)/ 50) + ClientHeight div 2;
  end;

begin
  if not Assigned(Map) then Exit;

  for I := 0 to Map.Roads.Count - 1 do
    with Map.Roads[I] as TRoad do begin
      if not actViewRoads.Checked and not Highlight then Continue;
      if Highlight then
        ptbMap.Canvas.Pen.Width := 2;
      if SelectedRoad = Map.Roads[I] then
        ptbMap.Canvas.Pen.Color := clMaroon
      else if Highlight then begin
        ptbMap.Canvas.Pen.Color := RGB(255, 128, 128)
      end else
        if Velocity <= 60 then
          ptbMap.Canvas.Pen.Color := clRed
        else if Velocity > 100 then
          ptbMap.Canvas.Pen.Color := clLime
        else
          ptbMap.Canvas.Pen.Color := clYellow;

      ptbMap.Canvas.MoveTo(CalcX(tFrom.x), CalcY(tFrom.Y));
      ptbMap.Canvas.LineTo(CalcX(tTo.X), CalcY(tTo.Y));

      ptbMap.Canvas.Pen.Width := 1;
      ptbMap.Canvas.Brush.Color := ptbMap.Canvas.Pen.Color;
      ptbMap.Canvas.Rectangle(CalcX(Center.X) - 3, CalcY(Center.Y) - 3, CalcX(Center.X) + 3, CalcY(Center.Y) + 3);

 {     ptbMap.Canvas.Brush.Style := bsClear;
      ptbMap.Canvas.TextOut(CalcX(Center.X) - ptbMap.Canvas.TextWidth(IntToStr(Length)),
        CalcY(Center.Y) - 7, IntToStr(Length));   }
    end;

  for I := 0 to Map.Towns.Count - 1 do
    with Map.Towns[I] as TTown do begin
      ptbMap.Canvas.Pen.Width := 1;
      if SelectedTown = Map.Towns[I] then begin
        ptbMap.Canvas.Pen.Color := clMaroon;
        ptbMap.Canvas.Brush.Color := clMaroon;
      end else begin
        ptbMap.Canvas.Pen.Color := clBlue;
        ptbMap.Canvas.Brush.Color := clBlue;
      end;
      ptbMap.Canvas.Ellipse(CalcX(X) - 5, CalcY(Y) - 5, CalcX(X) + 5, CalcY(Y) + 5);

      ptbMap.Canvas.Brush.Style := bsClear;
      ptbMap.Canvas.TextOut(CalcX(X) - ptbMap.Canvas.TextWidth(Name) div 2,
        CalcY(Y) - 20, Name);
//      ptbMap.Canvas.TextOut(CalcX(X) + 5, CalcY(Y) - ptbMap.Canvas.TextHeight('Wg') div 2, Format('%s', [Name]));
    end;
end;

procedure TfrmMain.scbHorizScroll(Sender: TObject; ScrollCode: TScrollCode;
  var ScrollPos: Integer);
begin
  ptbMap.Invalidate;
end;

procedure TfrmMain.actViewRoadsExecute(Sender: TObject);
begin
  actViewRoads.Checked := not actViewRoads.Checked;
  ptbMap.Invalidate;
end;

procedure TfrmMain.trbZoomChange(Sender: TObject);
begin
  ptbMap.Invalidate;
end;

procedure TfrmMain.ptbMapMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  nX, nY: Integer;
  I: Integer;
begin
// Prepocitame souradnice X, Y...
  nX := Round((X - ClientWidth div 2) / ((trbZoom.Position + 25) / 50)) + scbHoriz.Position;
  nY := Round((Y - Clientheight div 2) / ((trbZoom.Position + 25) / 50)) + scbVert.Position;

  //mesta
  for I := 0 to Map.Towns.Count - 1 do
    with Map.Towns[I] as TTown do begin
      if PtInRect(Rect(X - 5, Y - 5, X + 5, Y + 5), Point(nX, nY)) then begin
        SelectedTown := Map.Towns[I] as TTown;
        ptbMap.Invalidate;
        Exit;
      end;
    end;

  //cesty
  for I := 0 to Map.Roads.Count - 1 do
    with Map.Roads[I] as TRoad do begin
      if PtInRect(Rect(Center.X - 3, Center.Y - 3, Center.X + 3, Center.Y + 3), Point(nX, nY)) then begin
        if actViewRoads.Checked or Highlight then begin
          SelectedRoad := Map.Roads[I] as TRoad;
          ptbMap.Invalidate;
          Exit;
        end;
      end;
    end;
end;

procedure TfrmMain.ptbMapMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  nX, nY: Integer;
  I: Integer;
begin
  if SelectedTown = nil then Exit;
// Prepocitame souradnice X, Y...
  nX := Round((X - ClientWidth div 2) / ((trbZoom.Position + 25) / 50)) + scbHoriz.Position;
  nY := Round((Y - Clientheight div 2) / ((trbZoom.Position + 25) / 50)) + scbVert.Position;
  //mesta
  for I := 0 to Map.Towns.Count - 1 do
    with Map.Towns[I] as TTown do begin
      if PtInRect(Rect(X - 5, Y - 5, X + 5, Y + 5), Point(nX, nY)) then begin
        NajdiCestu(SelectedTown, Map.Towns[I] as TTown);
        Exit;
      end;
    end;
end;

procedure TfrmMain.NajdiCestu(tFrom, tTo: TTown);
var
  I: Integer;
  Town: TTown;
  ToTown: TTown;
  minTown: TTown;

  function GetRoadCost(Road: TRoad): Double;
  begin
    if mniKritLength.Checked then
      Result := Road.Length
    else if mniKritVelocity.Checked then
      Result := Road.Length / Road.Velocity
    else
      Result := Map.GetSpotreba(Road.Velocity) * Road.Length;
  end;

var
  Cost: Double;
begin
  if tFrom = tTo then Exit;
  PathFrom := tFrom;
  pathTo := tTo;
  stbMain.SimpleText := 'Hledam nejkratsi cestu...';

  // Dijkstruv algoritmus...
  for I := 0 to Map.Towns.Count - 1 do
    with Map.Towns[I] as TTown do begin
      DijkState := dsTemp;
      TownDist := MaxInt;
    end;

  tFrom.TownDist := 0;
  tFrom.DijkState := dsFinal;
  Town := tFrom;

  while tTo.DijkState <> dsFinal do begin
    for I := 0 to Map.Roads.Count - 1 do
      with Map.Roads[I] as TRoad do begin
        Cost := GetRoadCost(Map.Roads[I] as TRoad);
        ToTown := nil;
        if tFrom = Town then ToTown := tTo;
        if tTo = Town then ToTown := tFrom;
        if Assigned(ToTown) then
          if ToTown.TownDist > Town.TownDist + Cost then begin
            ToTown.TownDist := Town.TownDist + Cost;
            ToTown.FromTown := Town;
            ToTown.Road := Map.Roads[I] as TRoad;
          end;
      end;
    MinTown := nil;
    for I := 0 to Map.Towns.Count - 1 do
      with Map.Towns[I] as TTown do begin
        if DijkState = dsFinal then Continue;
        if (MinTown = nil) or (MinTown.TownDist > TownDist) then
          MinTown := Map.Towns[I] as TTown;
      end;
    if Assigned(MinTown) and (MinTown.TownDist <> MaxInt) then begin
      MinTown.DijkState := dsFinal;
      Town := MinTown;
    end else
      raise Exception.Create('Mezi mesty neexistuje cesta!!');
  end;

  for I := 0 to Map.Roads.Count - 1 do
    (Map.Roads[I] as TRoad).Highlight := False;

  Len := 0; Time := 0; palivo := 0;
  Town := tTo;
  while Town <> tFrom do begin
    Town.Road.Highlight := True;
    Len := Len + Town.Road.Length;
    Time := Time + Town.Road.Length / Town.Road.Velocity;
    palivo := palivo + Map.GetSpotreba(Town.Road.Velocity) * Town.Road.Length / 100;
    Town := Town.FromTown;
  end;

  FSelectedTown := nil;
  ptbMap.Invalidate;
  stbMain.SimpleText := Format('Nejkrat cesta z %s do %s ma delku %d, jzda by trvala %s a spotreba by byla %f',
    [PathFrom.Name, PathTo.Name, Round(Len), TimeToStr(Time / 24), palivo]);

  lblSelection.Caption := Format('Z %s do %s', [PathFrom.Name, PathTo.Name]);
  lblLength.Caption := IntToStr(Round(Len));
  lblTime.Caption := TimeToStr(Time / 24);
  lblLitru.Caption := FloatToStr(palivo);
end;

procedure TfrmMain.mniKritLengthClick(Sender: TObject);
begin
  (Sender as TMenuItem).Checked := True;
end;

procedure TfrmMain.actSavePathExecute(Sender: TObject);
var
  S: TStringList;
  Town: TTown;
begin
  if svdPath.Execute then begin
    S := TStringList.Create;;
    try
      S.Add('<?xml version="1.0" encoding="windows-1250" ?>');
      S.Add('<cesta>');
      S.Add(Format('<z>%s</z>', [PathFrom.Name]));
      S.Add(Format('<do>%s</do>', [PathTo.Name]));
      S.Add(Format('<vzdalenost>%d</vzdalenost>', [Round(Len)]));
      S.Add(Format('<dobaJizdy>%s</dobaJizdy>', [TimeToStr(Time / 24)]));
      S.Add(Format('<spotreba>%f</spotreba>', [palivo]));
      S.Add('</cesta>');

      Town := PathTo.FromTown;

      while Town <> PathFrom do begin
        S.Insert(4, Format('<pres>%s</pres>', [Town.Name]));
        Town := Town.FromTown;
      end;
      S.SaveToFile(svdPath.FileName);
    finally
      S.Free;
    end;
  end;
end;

procedure TfrmMain.actSavePathUpdate(Sender: TObject);
begin
  actSavePath.Enabled := Assigned(PathFrom) and Assigned(PathTo);
end;

procedure TfrmMain.actViewInfopanelExecute(Sender: TObject);
begin
  actViewInfopanel.Checked := not actViewInfopanel.Checked;
  pnlInfopanel.Visible := actViewInfopanel.Checked;
end;

end.
